/* inputd
 *
 * Copyright (C) 2006, 2007 Michael Hanselmann
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 */

#include <unistd.h>
#include <string.h>
#include <glib.h>
#include <glib/gprintf.h>

#include "common.h"
#include "brain.h"
#include "uinput.h"
#include "bitops.h"
#include "keys.h"

typedef struct {
    InputDevice *inputdev;
    struct input_event *event;
} BrainEvent;

static GList *queue;
static gboolean use_queue;
static gboolean wanted_key[KEY_MAX] = {
    [0 ... (KEY_MAX - 1)] = FALSE,
};

/* {{{ BrainEvent functions */
static BrainEvent *new_brain_event(InputDevice *inputdev,
    struct input_event *ev)
{
    BrainEvent *be = g_slice_new(BrainEvent);
    be->inputdev = inputdev;
    be->event = g_slice_new(struct input_event);
    memcpy(be->event, ev, sizeof(struct input_event));
    return be;
}

static void free_brain_event(BrainEvent *be)
{
    g_slice_free(struct input_event, be->event);
    g_slice_free(BrainEvent, be);
}
/* }}} */

/* {{{ Mouse emulation */
gboolean brain_need_mouse_emulation(InputDevice *inputdev,
    unsigned long *key_bits, guint len)
{
    struct work_data {
        unsigned long *bits;
        gboolean *result;
    };

    gboolean result = FALSE;
    struct work_data wdata = {
        .bits   = key_bits,
        .result = &result,
    };

    void work(gpointer data, gpointer user_data) {
        ConfigTranslation *cfgtr = data;
        struct work_data *wdata = user_data;
        unsigned long *bits = wdata->bits;
        gboolean *result = wdata->result;

        /* Check whether last key code of translation may come from this
         * device.
         */
        if (test_bit(CFGTR_LAST_KEY(cfgtr).def->code, bits)) {
            /* Target key a button? */
            if (cfgtr->target.def->flags & KEY_FLAG_BUTTON) {
                *result = TRUE;
            }
        }
    }

    g_list_foreach(config.translations, work, &wdata);

    return result;
}

void brain_determine_additional_bits(InputDevice *inputdev,
    guint16 type, unsigned long *bits, guint len)
{
    struct work_data {
        unsigned long *bits;
    };

    struct work_data wdata = {
        .bits = bits,
    };

    void key_work(gpointer data, gpointer user_data) {
        ConfigTranslation *cfgtr = data;
        struct work_data *wdata = user_data;
        unsigned long *bits = wdata->bits;

        /* Check whether last key code of translation may come from this
         * device.
         */
        if (test_bit(CFGTR_LAST_KEY(cfgtr).def->code, bits)) {
            /* If it can, activate target key bit */
            set_bit(cfgtr->target.def->code, bits);
        }
    }

    switch (type) {
        case 0:
            if (inputdev->mouse_emulation) {
                set_bit(EV_REL, bits);
            }
            break;

        case EV_KEY:
            g_list_foreach(config.translations, key_work, &wdata);

            if (inputdev->mouse_emulation) {
                set_bit(BTN_LEFT, bits);
            }
            break;

        case EV_REL:
            if (inputdev->mouse_emulation) {
                set_bit(REL_X, bits);
                set_bit(REL_Y, bits);
            }
            break;
    }
}
/* }}} */

/* {{{ Activate translation on queue */
static void build_modifier_events(ConfigTranslation *cfgtr,
    GList **mods_disable, GList **mods_enable)
{
    struct input_event new_ev;
    GList *cur;

    memset(&new_ev, 0, sizeof(new_ev));

    cur = devmanager_device_list();
    while (cur && cur->data) {
        InputDevice *inputdev = cur->data;
        gboolean sync = FALSE;
        guint i;

        for (i = 0; i < (cfgtr->key_count - 1); ++i) {
            if (inputdev->key[cfgtr->key[i].def->code].pressed) {
                new_ev.type = EV_KEY;
                new_ev.code = cfgtr->key[i].def->code;

                new_ev.value = 0;
                *mods_disable = g_list_append(*mods_disable,
                    new_brain_event(inputdev, &new_ev));

                new_ev.value = 1;
                *mods_enable = g_list_append(*mods_enable,
                    new_brain_event(inputdev, &new_ev));

                sync = TRUE;
            }
        }

        if (sync) {
            new_ev.type = EV_SYN;
            new_ev.code = 0;
            new_ev.value = 0;
            *mods_disable = g_list_append(*mods_disable,
                new_brain_event(inputdev, &new_ev));
            *mods_enable = g_list_append(*mods_enable,
                new_brain_event(inputdev, &new_ev));
        }

        cur = g_list_next(cur);
    }
}

static void apply_translation(ConfigTranslation* cfgtr)
{
    GList *mods_disable = NULL;
    GList *mods_enable = NULL;
    GList *cur;

    build_modifier_events(cfgtr, &mods_disable, &mods_enable);

    cur = g_list_first(queue);
    while (cur && cur->data) {
        BrainEvent *be = cur->data;
        struct input_event *ev = be->event;

        if (ev->type == EV_KEY &&
            CFGTR_LAST_KEY(cfgtr).def->code == ev->code) {
            ev->code = cfgtr->target.def->code;
            cfgtr->active = EV_PRESSED(ev);
            break;
        }

        cur = g_list_next(cur);
    }

    queue = g_list_concat(mods_disable, queue);
    queue = g_list_concat(queue, mods_enable);
}
/* }}} */

/* {{{ Check translations */
static void check_translations()
{
    GList *cur_tr;
    ConfigTranslation *apply = NULL;

    cur_tr = g_list_first(config.translations);
    while (cur_tr && cur_tr->data) {
        ConfigTranslation *cfgtr = cur_tr->data;
        gboolean all = TRUE;
        guint i;

        for (i = 0; i < cfgtr->key_count; i++) {
            GList *cur_ev;

#define curkey (cfgtr->key[i])

            cur_ev = g_list_first(queue);
            while (cur_ev && cur_ev->data) {
                BrainEvent *be = cur_ev->data;
                struct input_event *ev = be->event;

                if (ev->type == EV_KEY &&
                    curkey.def->code == ev->code) {
                    curkey.pressed = EV_PRESSED(ev);
                }

                cur_ev = g_list_next(cur_ev);
            }

            all = all && curkey.pressed;

#undef curkey
        }

        if (all && !cfgtr->active) {
            /* Start translation */
            apply = cfgtr;

        } else if (!all && cfgtr->active) {
            /* End translation */
            apply = cfgtr;
        }

        cur_tr = g_list_next(cur_tr);
    }

    if (apply) {
        apply_translation(apply);
    }
}
/* }}} */

/* {{{ Send queued events */
static void send_queue_events()
{
    GList *cur;

    cur = g_list_first(queue);
    while (cur && cur->data) {
        BrainEvent *be = cur->data;

        uinput_write_event2(be->inputdev, be->event);
        free_brain_event(be);

        cur = g_list_next(cur);
    }

    g_list_free(queue);
    queue = NULL;
}
/* }}} */

/* {{{ Add event to queue */
void brain_submit_event(InputDevice *inputdev, struct input_event *ev)
{
    inputdev->key[ev->code].pressed = (ev->value > 0);

    if (!use_queue && ev->type == EV_KEY) {
        /* Only use the queue if needed */
        use_queue = use_queue || wanted_key[ev->code];
    }

    if (!use_queue) {
        if (queue) {
            send_queue_events();
        }

        uinput_write_event2(inputdev, ev);

        if (ev->type == EV_SYN) {
            use_queue = FALSE;
        }
    } else {
        queue = g_list_append(queue, new_brain_event(inputdev, ev));

        if (ev->type == EV_SYN) {
            check_translations();
            send_queue_events();
            use_queue = FALSE;
        }
    }
}
/* }}} */

/* {{{ Init/Exit */
gboolean brain_init()
{
    void build_wanted_keys(gpointer data, gpointer user_data) {
        ConfigTranslation *cfgtr = data;
        int i;

        for (i = 0; i < cfgtr->key_count; i++) {
            wanted_key[cfgtr->key[i].def->code] = TRUE;
        }
    }

    g_list_foreach(config.translations, build_wanted_keys, NULL);

    queue = NULL;
    use_queue = FALSE;

    return TRUE;
}

void brain_exit()
{
}
/* }}} */

/* vim: set sw=4 sts=4 et foldmethod=marker : */
