%error-verbose
%locations

%{
#include <glib.h>
#include <stdio.h>
#include <string.h>

#include "common.h"
#include "config.h"
#include "keys.h"

int yylex();
int yyget_lineno(void);

enum type {
    T_STRING,
    T_NUMERIC,
    T_TRANSLATIONS,
};

struct variable {
    enum type type;
    union {
        gchar **str;
        gint *num;
        GList **list;
    };
};

struct value {
    enum type type;
    union {
        gchar *str;
        gint num;
        GList *list;
    };
};

void yyerror(const char *msg)
{
    g_critical("Line %d: %s", yyget_lineno(), msg);
}

%}

%union {
    char *cptr;
    void *ptr;
    ConfigTranslation *cfgtr;
    struct variable *variable;
    struct value *value;
    GList *list;
    KeyDefinition *kd;
    gint number;
}

%token <cptr> STRING ID
%token <number> NUMBER

%type <variable> variable;
%type <value> value;
%type <list> keycombo translations;
%type <kd> key;
%type <cfgtr> translation;

%left '+'

%destructor { g_free($$); } STRING variable value;
%destructor { free_config_translation($$); } translation;

%%

start:
    stmts
    ;

stmts:
    /* empty */
    | assign ';' stmts
    ;

variable:
    ID {
        $$ = g_new0(typeof(*$$), 1);

        if (strcmp($1, "uinput_dev") == 0) {
            $$->type = T_STRING;
            $$->str = &config.uinput_dev;

        } else if (strcmp($1, "input_dev_dir") == 0) {
            $$->type = T_STRING;
            $$->str = &config.input_dev_dir;

        } else if (strcmp($1, "verbosity") == 0) {
            $$->type = T_NUMERIC;
            $$->num = &config.verbosity;

        } else if (strcmp($1, "translations") == 0) {
            $$->type = T_TRANSLATIONS;
            $$->list = &config.translations;

        } else {
            g_free($$);
            yyerror("Unknown variable");
            YYERROR;
        }
    }
    ;

value:
    STRING {
        $$ = g_new0(struct value, 1);
        $$->type = T_STRING;
        $$->str = $1;
    }
    | NUMBER {
        $$ = g_new0(struct value, 1);
        $$->type = T_NUMERIC;
        $$->num = $1;
    }
    | '{' translations '}' {
        $$ = g_new0(struct value, 1);
        $$->type = T_TRANSLATIONS;
        $$->list = $2;
    }
    ;

assign:
    variable '=' value {
        if ($1->type != $3->type) {
            g_free($1);
            g_free($3);
            yyerror("Invalid value");
            YYERROR;
        }

        switch ($1->type) {
        case T_STRING:
            *($1->str) = g_strdup($3->str);
            g_free($3->str);
            break;
        case T_NUMERIC:
            *($1->num) = $3->num;
            break;
        case T_TRANSLATIONS:
            set_translations($1->list, $3->list);
            break;
        }

        g_free($1);
        g_free($3);
    }
    ;

key:
    ID {
        $$ = key_by_name($1);
        if ($$ == NULL) {
            yyerror("Unknown key");
            YYERROR;
        }
    }
    ;

keycombo:
    key                     { $$ = g_list_append(NULL, $1); }
    | keycombo '+' key      { $$ = g_list_append($1, $3); }
    ;

translation:
    keycombo '=' key {
        $$ = new_config_translation($1, $3);
        g_list_free($1);
    }
    ;

translations:
    /* empty */                     { $$ = NULL; }
    | translations translation ';'  { $$ = g_list_append($1, $2); }
    ;

%%

/* vim: set sw=4 sts=4 et foldmethod=marker : */
