/* inputd
 *
 * Copyright (C) 2006, 2007 Michael Hanselmann
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 */

#include "../config.h"

#include <glib.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <stdlib.h>
#include <errno.h>
#include <stdio.h>
#include <string.h>

#include "common.h"
#include "config.h"
#include "keys.h"
#include "utils.h"

Config config;

static Config defconfig = {
    .verbosity = 0,
    .uinput_dev = "/dev/misc/uinput",
    .input_dev_dir = "/dev/input",
};

/* {{{ ConfigTranslation functions */
ConfigTranslation *new_config_translation(GList *key_defs,
    KeyDefinition *target_def)
{
    ConfigTranslation *cfgtr = g_new0(ConfigTranslation, 1);
    guint count = g_list_length(key_defs);
    guint i;
    GList *cur;

    cfgtr->target.def = target_def;
    cfgtr->key_count = count;
    cfgtr->key = g_new0(ConfigKey, cfgtr->key_count);

    i = 0;
    cur = g_list_first(key_defs);
    while (cur && cur->data && i < cfgtr->key_count) {
        KeyDefinition *kd = cur->data;
        cfgtr->key[i++].def = kd;
        cur = g_list_next(cur);
    }

    return cfgtr;
}

void set_translations(GList **to, GList *translations)
{
    void work(gpointer data, gpointer user_data) {
        free_config_translation(data);
    }

    g_list_foreach(*to, work, NULL);
    g_list_free(*to);

    *to = translations;
}

void free_config_translation(ConfigTranslation *cfgtr)
{
    g_free(cfgtr->key);
    g_free(cfgtr);
}
/* }}} */

/* {{{ Command line */
struct CmdLineOptions {
    gchar *config_file;
    guint verbosity;
    gboolean show_version;
    guint quit_timeout;
};

static struct CmdLineOptions cmdline_opts = {
    .config_file = NULL,
    .verbosity = 0,
    .show_version = FALSE,
};

static gboolean parse_verbosity(const gchar *option_name, const gchar *value,
        gpointer data, GError **error)
{
    if (value) {
        cmdline_opts.verbosity = strtol(value, NULL, 10);
        if (errno == ERANGE) {
            cmdline_opts.verbosity = 0;
        }
    } else {
        cmdline_opts.verbosity++;
    }

    return TRUE;
}

static GOptionEntry cmdline_options_desc[] = {
    { "config-file", 'c',
      G_OPTION_FLAG_FILENAME, G_OPTION_ARG_FILENAME,
      &cmdline_opts.config_file, "Config file", "file" },
    { "verbosity", 'v',
      G_OPTION_FLAG_FILENAME | G_OPTION_FLAG_OPTIONAL_ARG,
      G_OPTION_ARG_CALLBACK,
      parse_verbosity, "Verbosity (repeat for higher verbosity)", NULL },
    { "quit-timeout", 0,
      G_OPTION_FLAG_HIDDEN, G_OPTION_ARG_INT,
      &cmdline_opts.quit_timeout, "Quit after n seconds", "n" },
    { "version", 0,
      G_OPTION_FLAG_NO_ARG, G_OPTION_ARG_NONE,
      &cmdline_opts.show_version, "Show version", NULL },
    { NULL }
};

static void show_version()
{
    g_print("%s\n",
        PACKAGE_STRING "\n"
        "Copyright (C) 2006, 2007 Michael Hanselmann\n"
        "This program is free software; you can redistribute it and/or modify\n"
        "it under the terms of the GNU General Public License as published by\n"
        "the Free Software Foundation; either version 2 of the License, or\n"
        "(at your option) any later version.");
}

/* Called before config_init */
gboolean config_parse_cmdline(int argc, char *argv[])
{
    GOptionContext *context;
    gboolean result = TRUE;
    GError *error = NULL;

    context = g_option_context_new(NULL);
    g_option_context_add_main_entries(context, cmdline_options_desc, NULL);
    g_option_context_parse(context, &argc, &argv, &error);
    if (error) {
        g_critical(error->message);
        g_error_free(error);

        result = FALSE;
    }
    g_option_context_free(context);

    if (cmdline_opts.show_version) {
        show_version();
        result = FALSE;
    }

    return result;
}
/* }}} */

/* {{{ Config file */
extern void yyrestart(FILE *);
extern int yyparse(void);
extern void yylex_destroy(void);

static gboolean parse_config(const char *file)
{
    FILE *cfg;
    int res;

    cfg = fopen(file, "r");
    if (cfg == NULL) {
        g_critical("%s: %s", file, strerror(errno));
        return FALSE;
    }

    yyrestart(cfg);
    res = yyparse();
    fclose(cfg);
    yylex_destroy();

    return (res == 0);
}
/* }}} */

/* {{{ Init/Exit */
gboolean config_init()
{
    char *config_file;

    /* Determine which config file is to be used */
    if (cmdline_opts.config_file) {
        config_file = cmdline_opts.config_file;
    } else {
        config_file = SYSCONFDIR "/inputd.conf";
    }

    /* Try to parse config file */
    if (!parse_config(config_file)) {
        g_critical("Failed to parse config");
        return FALSE;
    }

    /* Override unset options using default values */
    if (!config.uinput_dev) {
        config.uinput_dev = g_strdup(defconfig.uinput_dev);
    }
    if (!config.input_dev_dir) {
        config.input_dev_dir = g_strdup(defconfig.input_dev_dir);
    }

    /* Options that are only available on the command line */
    config.quit_timeout = cmdline_opts.quit_timeout;

    /* Override config options using command line arguments */
    if (cmdline_opts.verbosity) {
        config.verbosity = cmdline_opts.verbosity;
    }

    if (config.verbosity >= VERBOSITY_DEBUG) {
        dump_config();
    }

    return TRUE;
}

void config_exit()
{
    /* Free memory */
    g_free(config.uinput_dev);
    g_free(config.input_dev_dir);

    set_translations(&config.translations, NULL);
}
/* }}} */

/* vim: set sw=4 sts=4 et foldmethod=marker : */
