/* inputd
 *
 * Copyright (C) 2006, 2007 Michael Hanselmann
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 */

#include <glib.h>
#include <fcntl.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>

#include "common.h"
#include "devmanager.h"
#include "eventsource.h"
#include "uinput.h"
#include "inotify.h"

static Inotify *dev_inotify = NULL;
static GList *devices = NULL;

static void print_devices()
{
    GList *cur;
    guint i;

    g_message("-- Current devices -----");

    for (i = 0, cur = g_list_first(devices);
         cur != NULL;
         cur = g_list_next(cur)) {

        InputDevice *dev = cur->data;

        g_message("%d: %s", i, dev->path);

        i++;
    }

    g_message("------------------------");
}

static gint _find_by_path_cmp(gconstpointer _dev, gconstpointer _path)
{
    const InputDevice *dev = _dev;
    const char *path = _path;

    return strcmp(dev->path, path) == 0?0:-1;
}

static InputDevice *find_device(InputDevice *inputdev)
{
    GList *found;

    gint _find_device_compare(gconstpointer _a, gconstpointer _b) {
        const InputDevice *a = _a;
        const InputDevice *b = _b;

        if (strcmp(a->name, b->name) == 0 &&
            strcmp(a->phys, b->phys) == 0 &&
            strcmp(a->path, b->path) == 0) {
            return 0;
        } else {
            return -1;
        }
    }

    found = g_list_find_custom(devices, inputdev, _find_device_compare);

    if (found) {
        return found->data;
    } else {
        return NULL;
    }
}

static void try_device(const char *path)
{
    InputDevice *inputdev, *found;
    int fd;

    fd = open(path, O_RDWR);
    if (fd < 0) {
        return;
    }

    /* Allocate device */
    inputdev = g_new0(InputDevice, 1);

    /* Set device */
    g_strlcpy(inputdev->path, path, sizeof(inputdev->path));

    /* Retrieve name */
    if (ioctl(fd, EVIOCGNAME(sizeof(inputdev->name)), inputdev->name) < 0) {
        g_stpcpy(inputdev->name, "");
    }

    /* Check wether it's one of our uinput devices */
    if (g_str_has_prefix(inputdev->name, INPUTD_DEVICE_PREFIX)) {
        goto out_free;
    }

    /* Retrieve phys */
    if (ioctl(fd, EVIOCGPHYS(sizeof(inputdev->phys)), inputdev->phys) < 0) {
        g_stpcpy(inputdev->phys, "");
    }

    /* Retrieve device ID */
    if (ioctl(fd, EVIOCGID, &inputdev->id) < 0) {
        g_critical("ioctl(EVIOCGID) failed: %s", strerror(errno));
        goto out_free;
    }

    /* Try to find existing device */
    found = find_device(inputdev);
    if (found) {
        goto out_free;
    }

    if (config.verbosity >= VERBOSITY_INFO) {
        g_message("Adding %s", inputdev->name);
    }

    /* Create event source */
    event_source_create(inputdev, fd);

    if (!event_source_reset(inputdev)) {
        g_critical("Resetting event source failed");
        goto out_evsource;
    }

    if (!event_source_grab(inputdev, TRUE)) {
        g_critical("Grabbing input device failed");
        goto out_evsource;
    }

    if (!uinput_create(inputdev)) {
        g_critical("Uinput device creation failed");
        goto out_evsource;
    }

    event_source_add_watch(inputdev);

    devices = g_list_append(devices, inputdev);

    return;

out_evsource:
    event_source_destroy(inputdev);

out_free:
    g_free(inputdev);

    close(fd);

    return;
}

static void free_device(InputDevice *inputdev)
{
    if (config.verbosity >= VERBOSITY_INFO) {
        g_message("Removing %s", inputdev->name);
    }

    uinput_destroy(inputdev);
    event_source_destroy(inputdev);

    g_free(inputdev);
}

static void unplug_device(const char *path)
{
    GList *found;

    found = g_list_find_custom(devices, path, _find_by_path_cmp);
    if (found != NULL) {
        free_device(found->data);

        devices = g_list_delete_link(devices, found);
    }
}

static gboolean inotify_callback(Inotify *inot, const char *name,
        int wd, unsigned int mask, unsigned int cookie)
{
    if (name && g_str_has_prefix(name, "event")) {
        char *path = g_build_filename(config.input_dev_dir, name, NULL);

        if (mask & IN_CREATE) {
            try_device(path);

        } else if (mask & IN_DELETE) {
            unplug_device(path);
        }

        if (config.verbosity >= VERBOSITY_DEBUG) {
            g_message("Detected change on %s", path);
            print_devices();
        }

        g_free(path);
    }

    return TRUE;
}

void devmanager_foreach_dev(DeviceCallback cb, gpointer user_data)
{
    struct work_data {
        DeviceCallback *cb;
        gpointer user_data;
    };
    struct work_data wdata = {
        .cb = cb,
        .user_data = user_data,
    };

    void work(gpointer data, gpointer user_data) {
        struct work_data *wdata = user_data;
        wdata->cb(data, wdata->user_data);
    }

    g_list_foreach(devices, work, &wdata);
}

GList *devmanager_device_list()
{
    return devices;
}

static void scan_all_devices()
{
    guint i;

    for (i = 0; i < 32; ++i) {
        char path[sizeof(((InputDevice*)0)->path)];

        g_snprintf(path, sizeof(path),
            "%s/event%d", config.input_dev_dir, i);

        try_device(path);
    }

    if (config.verbosity >= VERBOSITY_DEBUG) {
        print_devices();
    }
}

gboolean devmanager_init()
{
    dev_inotify = my_inotify_open();
    if (!dev_inotify) {
        return FALSE;
    }

    my_inotify_set_callback(dev_inotify, inotify_callback);

    if (my_inotify_add_watch(dev_inotify, config.input_dev_dir,
            IN_CREATE | IN_DELETE) < 0) {
        my_inotify_close(dev_inotify);

        return FALSE;
    }

    scan_all_devices();

    return TRUE;
}

void devmanager_exit()
{
    GList *cur;

    if (dev_inotify) {
        my_inotify_close(dev_inotify);
    }

    /* Release all devices */
    for (cur = g_list_first(devices); cur != NULL; cur = g_list_next(cur)) {
        free_device(cur->data);
    }
}

/* vim: set sw=4 sts=4 et foldmethod=marker : */
