/* inputd
 *
 * Copyright (C) 2006, 2007 Michael Hanselmann
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 */

#include <errno.h>
#include <unistd.h>
#include <string.h>

#include "common.h"
#include "brain.h"
#include "eventsource.h"
#include "bitops.h"

static gboolean _event_source_watch(GIOChannel *io,
        GIOCondition condition, gpointer data)
{
    InputDevice *inputdev = data;
    gboolean result = FALSE;

    if (condition & G_IO_IN) {
        struct input_event buf[16] = {{{0}}};
        ssize_t len;
        int fd;

        fd = g_io_channel_unix_get_fd(io);

        len = read(fd, buf, sizeof(buf));
        if (len < 0) {
            /* error */

        } else if (len >= sizeof(buf[0])) {
            guint count = MIN(len / sizeof(buf[0]), ARRAY_SIZE(buf));
            guint i;

            for (i = 0; i < count; i++) {
                brain_submit_event(inputdev, &buf[i]);
            }

            result = TRUE;
        }
    }

    return result;
}

gboolean event_source_reset(InputDevice *inputdev)
{
    int fd = g_io_channel_unix_get_fd(inputdev->evsource);
    unsigned long bits[KEY_MAX] = { 0 };
    int i;

    if (ioctl(fd, EVIOCGKEY(sizeof(bits)), bits) < 0) {
        g_critical("ioctl(EVIOCGKEY) failed: %s", strerror(errno));

        return FALSE;
    } 

    for (i = 0; i < KEY_MAX; ++i) {
        if (test_bit(i, bits)) {
            event_source_write_event(inputdev, EV_KEY, i, 0);
        }
    }

    return TRUE;
}

gboolean event_source_grab(InputDevice *inputdev, gboolean grab)
{
    int fd = g_io_channel_unix_get_fd(inputdev->evsource);

    if (ioctl(fd, EVIOCGRAB, (grab?1:0))) {
        g_critical("ioctl(EVIOCGRAB) failed: %s", strerror(errno));

        return FALSE;
    }

    return TRUE;
}

gboolean event_source_write_event(InputDevice *inputdev, guint16 type,
        guint16 code, gint32 value)
{
    int fd = g_io_channel_unix_get_fd(inputdev->evsource);
    struct input_event ev = {
        .type = type,
        .code = code,
        .value = value,
    };

    gettimeofday(&ev.time, NULL);

    return (write(fd, &ev, sizeof(struct input_event)) == sizeof(struct input_event));
}

void event_source_create(InputDevice *inputdev, int fd)
{
    inputdev->evsource = g_io_channel_unix_new(fd);

    /* Set options on IO channel */
    g_io_channel_set_close_on_unref(inputdev->evsource, TRUE);
    g_io_channel_set_encoding(inputdev->evsource, NULL, NULL);
}

void event_source_add_watch(InputDevice *inputdev)
{
    /* Add watch source */
    inputdev->evsource_watch = g_io_add_watch_full(
        inputdev->evsource,
        G_PRIORITY_DEFAULT,
        G_IO_IN | G_IO_ERR | G_IO_HUP,
        _event_source_watch, inputdev,
        NULL);
}

void event_source_destroy(InputDevice *inputdev)
{
    if (inputdev->evsource_watch) {
        g_source_remove(inputdev->evsource_watch);
        inputdev->evsource_watch = 0;
    }

    if (inputdev->evsource) {
        g_io_channel_unref(inputdev->evsource);
        inputdev->evsource = NULL;
    }
}

/* vim: set sw=4 sts=4 et foldmethod=marker : */
