/* inputd
 *
 * Copyright (C) 2006, 2007 Michael Hanselmann
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 */

#include <glib.h>
#include <errno.h>
#include <string.h>

#include "inotify.h"

/* Opens an IOChannel for inotify */
Inotify *my_inotify_open(void)
{
    GIOChannel *inot;
    GError *error = NULL;
    int fd;

    fd = inotify_init();
    if (fd < 0) {
        int tmp = errno;

        g_critical("inotify_init failed: %s", strerror(errno));

        if (tmp == ENOSYS) {
            g_error("Inotify not supported! You need Linux 2.6.13 or later "
                "with CONFIG_INOTIFY enabled. Your glibc needs support for "
                "it, too.");
        }

        return NULL;
    }

    inot = g_io_channel_unix_new(fd);

    g_io_channel_set_close_on_unref(inot, TRUE);
    g_io_channel_set_encoding(inot, NULL, NULL);

    g_io_channel_set_flags(inot, G_IO_FLAG_NONBLOCK, &error);
    if (error) {
        g_warning(error->message);
        g_error_free(error);
    }

    return inot;
}

/*
 * Close the GIOChannel
 */
void my_inotify_close(Inotify *inot)
{
    g_io_channel_close(inot);
}

/*
 * Add an inotify watch on the object "name" to the open inotify instance
 * associated with "gio". The user may do this any number of times, even on the
 * same device instance.
 */
int my_inotify_add_watch(Inotify *inot, const char *path, guint32 mask)
{
    int fd = g_io_channel_unix_get_fd(inot);
    int wd;

    wd = inotify_add_watch(fd, path, mask);
    if (wd < 0) {
        g_warning("inotify_add_watch failed: %s", strerror(errno));
    }

    return wd;
}

gboolean my_inotify_rm_watch(Inotify *inot, const int wd)
{
    int fd = g_io_channel_unix_get_fd(inot);
    int ret;

    ret = inotify_rm_watch(fd, wd);
    if (wd < 0) {
        g_warning("inotify_rm_watch failed: %s", strerror(errno));

        return FALSE;
    }

    return TRUE;
}

/*
 * Our internal GIOChannel G_IO_IN callback.  Slurps as many events as are
 * available and calls the user's callback (given as "data") for each event.
 * If any invocations of the user's callback return FALSE, so do we,
 * terminating this watch. Otherwise, we return TRUE.
 */
static gboolean __inotify_handle_cb(GIOChannel *io, GIOCondition condition,
        gpointer data)
{
    Inotify *inot = io;
    InotifyCb callback = data;
    char buf[sizeof(struct inotify_event) * 32];
    GError *error = NULL;
    GIOStatus status;
    gsize len, i = 0;

    status = g_io_channel_read_chars(io, buf, sizeof(buf), &len, &error);
    if (status != G_IO_STATUS_NORMAL) {
        if (error) {
            g_warning(error->message);
            g_error_free(error);
        }

        return FALSE;
    }

    while (i < len) {
        struct inotify_event *event = (struct inotify_event *)&buf[i];

        if (!callback(inot, event->name, event->wd, event->mask, event->cookie)) {
            return FALSE;
        }

        i += sizeof(struct inotify_event) + event->len;
    }

    return TRUE;
}

/*
 * Associate a user InotifyCb callback with the given GIOChannel. This is
 * normally done once.
 */
void my_inotify_set_callback(Inotify *inot, InotifyCb callback)
{
    g_io_add_watch(inot, G_IO_IN, __inotify_handle_cb, callback);
}

/* vim: set sw=4 sts=4 et foldmethod=marker : */
