/* inputd
 *
 * Copyright (C) 2006, 2007 Michael Hanselmann
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 */

#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <glib.h>
#include <glib/gprintf.h>

#include "common.h"
#include "eventsource.h"
#include "uinput.h"
#include "brain.h"
#include "bitops.h"
#include "utils.h"

static guint32 cookie_random = 0;
static guint32 cookie_count = 0;

static gboolean set_event_bits(InputDevice *inputdev)
{
    int fd_evsource = g_io_channel_unix_get_fd(inputdev->evsource);
    int fd_uinput = g_io_channel_unix_get_fd(inputdev->uinput);
    unsigned long ev_bits[NBITS(EV_MAX)] = { 0 };
    unsigned long key_bits[NBITS(KEY_MAX)] = { 0 };
    unsigned long rel_bits[NBITS(REL_MAX)] = { 0 };
    unsigned long abs_bits[NBITS(ABS_MAX)] = { 0 };
    unsigned long msc_bits[NBITS(MSC_MAX)] = { 0 };
    unsigned long led_bits[NBITS(LED_MAX)] = { 0 };
    unsigned long snd_bits[NBITS(SND_MAX)] = { 0 };
    unsigned long ff_bits[NBITS(FF_MAX)] = { 0 };
    unsigned long sw_bits[NBITS(SW_MAX)] = { 0 };

#define GET_BITS(_evtype, _bits) do { \
    if (ioctl(fd_evsource, EVIOCGBIT(_evtype, sizeof(_bits)), _bits) < 0) { \
        g_critical("ioctl(EVIOCGBIT(" #_evtype ")) failed: %s", \
            strerror(errno)); \
        return FALSE; \
    } \
} while (0);

    GET_BITS(0, ev_bits);
    GET_BITS(EV_KEY, key_bits);
    GET_BITS(EV_REL, rel_bits);
    GET_BITS(EV_ABS, abs_bits);
    GET_BITS(EV_MSC, msc_bits);
    GET_BITS(EV_LED, led_bits);
    GET_BITS(EV_SND, snd_bits);
    GET_BITS(EV_FF, ff_bits);
    GET_BITS(EV_SW, sw_bits);

#undef GET_BITS

    inputdev->mouse_emulation =
        brain_need_mouse_emulation(inputdev, key_bits, sizeof(key_bits));

    brain_determine_additional_bits(inputdev, 0,
        ev_bits, sizeof(ev_bits));
    brain_determine_additional_bits(inputdev, EV_KEY,
        key_bits, sizeof(key_bits));
    brain_determine_additional_bits(inputdev, EV_REL,
        rel_bits, sizeof(rel_bits));
    brain_determine_additional_bits(inputdev, EV_ABS,
        abs_bits, sizeof(abs_bits));
    brain_determine_additional_bits(inputdev, EV_MSC,
        msc_bits, sizeof(msc_bits));
    brain_determine_additional_bits(inputdev, EV_LED,
        led_bits, sizeof(led_bits));
    brain_determine_additional_bits(inputdev, EV_SND,
        snd_bits, sizeof(snd_bits));
    brain_determine_additional_bits(inputdev, EV_FF,
        ff_bits, sizeof(ff_bits));
    brain_determine_additional_bits(inputdev, EV_SW,
        sw_bits, sizeof(sw_bits));

#define SET_BITS(_evtype, _ctluinput, _max, _bits) do { \
    int i; \
    \
    for (i = 0; i < _max; ++i) { \
        if (test_bit(i, _bits) && ioctl(fd_uinput, _ctluinput, i) < 0) { \
            g_critical("ioctl(" #_ctluinput ") failed: %s", strerror(errno)); \
            return FALSE; \
        } \
    } \
} while (0);

    SET_BITS(0, UI_SET_EVBIT, EV_MAX, ev_bits);
    SET_BITS(EV_KEY, UI_SET_KEYBIT, KEY_MAX, key_bits);
    SET_BITS(EV_REL, UI_SET_RELBIT, REL_MAX, rel_bits);
    SET_BITS(EV_ABS, UI_SET_ABSBIT, ABS_MAX, abs_bits);
    SET_BITS(EV_MSC, UI_SET_MSCBIT, MSC_MAX, msc_bits);
    SET_BITS(EV_LED, UI_SET_LEDBIT, LED_MAX, led_bits);
    SET_BITS(EV_SND, UI_SET_SNDBIT, SND_MAX, snd_bits);
    SET_BITS(EV_FF, UI_SET_FFBIT, FF_MAX, ff_bits);
    SET_BITS(EV_SW, UI_SET_SWBIT, SW_MAX, sw_bits);

#undef SET_BITS

    return TRUE;
}

static gboolean set_abs_axes(InputDevice *inputdev,
        struct uinput_user_dev *device)
{
    int fd_evsource, i;

    fd_evsource = g_io_channel_unix_get_fd(inputdev->evsource);

    for (i = 0; i < ABS_MAX; ++i) {
        struct input_absinfo abs;

        /* Retrieve values */
        if (ioctl(fd_evsource, EVIOCGABS(i), &abs) < 0) { \
            g_critical("ioctl(EVIOCGABS) failed: %s", strerror(errno));
            return FALSE;
        }

        /* Set values on uinput device */
        device->absmin[i] = abs.minimum;
        device->absmax[i] = abs.maximum;
        device->absfuzz[i] = abs.fuzz;
        device->absflat[i] = abs.flat;
    }

    return TRUE;
}

gboolean uinput_create(InputDevice *inputdev)
{
    struct uinput_user_dev device;
    GError *error = NULL;
    gchar *newname;
    int fd;

    inputdev->uinput = g_io_channel_new_file(config.uinput_dev, "r+", &error);
    if (error) {
        char *msg = g_strconcat(config.uinput_dev, ": ", error->message, NULL);
        g_warning(msg);
        g_error_free(error);
        g_free(msg);

        return FALSE;
    }

    g_io_channel_set_close_on_unref(inputdev->uinput, TRUE);
    g_io_channel_set_encoding(inputdev->uinput, NULL, NULL);
    g_io_channel_set_buffered(inputdev->uinput, FALSE);

    fd = g_io_channel_unix_get_fd(inputdev->uinput);

    /* Reset device structure */
    memset(&device, 0, sizeof(device));

    /* Create new device name */
    newname = g_strconcat(INPUTD_DEVICE_PREFIX, inputdev->name, NULL);
    g_strlcpy(device.name, newname, sizeof(device.name));
    g_free(newname);

    /* Copy device ID values */
    memcpy(&device.id, &inputdev->id, sizeof(device.id));

    /* Set of absolute axes */
    if (!set_abs_axes(inputdev, &device)) {
        g_io_channel_unref(inputdev->uinput);

        return FALSE;
    }

    /* Set device info */
    if (write(fd, &device, sizeof(device)) < sizeof(device)) {
        g_critical("write failed: %s", strerror(errno));

        g_io_channel_unref(inputdev->uinput);

        return FALSE;
    }

    /* Enable event bits */
    if (!set_event_bits(inputdev)) {
        g_io_channel_unref(inputdev->uinput);

        return FALSE;
    }

    /* Build cookie */
    g_snprintf(inputdev->uinput_cookie, sizeof(inputdev->uinput_cookie),
        "inputd:%08x%08x%04x%04x%d",
        GPOINTER_TO_INT(inputdev) ^ cookie_random,
        g_str_hash(inputdev->name),
        inputdev->id.vendor,
        inputdev->id.product,
        ++cookie_count
    );

    /* Set cookie */
    ioctl(fd, UI_SET_PHYS, inputdev->uinput_cookie);

    /* Create uinput device */
    if (ioctl(fd, UI_DEV_CREATE, NULL)) {
        g_critical("ioctl(UI_DEV_CREATE) failed: %s", strerror(errno));

        g_io_channel_unref(inputdev->uinput);

        return FALSE;
    }

    return TRUE;
}

void uinput_destroy(InputDevice *inputdev)
{
    int i;

    /* Reset pressed keys */
    for (i = 0; i < KEY_MAX; i++) {
        if (inputdev->key[i].pressed) {
            uinput_write_event(inputdev, EV_KEY, i, 0);
        }
    }

    event_source_reset(inputdev);

    uinput_write_sync(inputdev);
    event_source_write_sync(inputdev);

    g_io_channel_unref(inputdev->uinput);

    inputdev->uinput = NULL;
}

gboolean uinput_write_event(InputDevice *inputdev,
        guint16 type, guint16 code, gint32 value)
{
    int fd = g_io_channel_unix_get_fd(inputdev->uinput);
    struct input_event ev = {
        .type = type,
        .code = code,
        .value = value,
    };

    gettimeofday(&ev.time, NULL);

    if (config.verbosity >= VERBOSITY_DEBUG) {
        dump_event(&ev);
    }

    return (write(fd, &ev, sizeof(struct input_event)) ==
            sizeof(struct input_event));
}

gboolean uinput_init()
{
    cookie_random = g_random_int();

    return TRUE;
}

void uinput_exit()
{
}

/* vim: set sw=4 sts=4 et foldmethod=marker : */
